@@ -1,14 +1,19 @@
+import json
 import logging
 import os
 import tempfile
+import subprocess
 import shutil
 from datetime import datetime, timedelta
 import pytest
 import pytz
+import sqlite3
 
 from ..client import MerginClient, ClientError, MerginProject, LoginError, decode_token_data, TokenError
 from ..client_push import push_project_async, push_project_cancel
-from ..utils import generate_checksum
+from ..utils import generate_checksum, get_versions_with_file_changes
+from ..merginproject import pygeodiff
+
 
 SERVER_URL = os.environ.get('TEST_MERGIN_URL')
 API_USER = os.environ.get('TEST_API_USERNAME')
@@ -20,10 +25,6 @@ TEST_DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test_
 CHANGED_SCHEMA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'modified_schema')
 
 
-def toggle_geodiff(enabled):
-    os.environ['GEODIFF_ENABLED'] = str(enabled)
-
-
 @pytest.fixture(scope='function')
 def mc():
     return create_client(API_USER, USER_PWD)
@@ -282,19 +283,9 @@ def test_ignore_files(mc):
     assert not next((f for f in project_info['files'] if f['path'] == '.directory'), None)
 
 
-# (diffs size limit, push geodiff enabled, pull geodiff enabled)
-diff_test_scenarios = [
-    (True, True),
-    (True, False),
-    (False, True),
-    (False, False),
-]
+def test_sync_diff(mc):
 
-
-@pytest.mark.parametrize("push_geodiff_enabled, pull_geodiff_enabled", diff_test_scenarios)
-def test_sync_diff(mc, push_geodiff_enabled, pull_geodiff_enabled):
-
-    test_project = f'test_sync_diff_push{int(push_geodiff_enabled)}_pull{int(pull_geodiff_enabled)}'
+    test_project = f'test_sync_diff'
     project = API_USER + '/' + test_project
     project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir for updates
     project_dir_2 = os.path.join(TMP_DIR, test_project + '_2')  # concurrent project dir with no changes
@@ -302,32 +293,24 @@ def test_sync_diff(mc, push_geodiff_enabled, pull_geodiff_enabled):
 
     cleanup(mc, project, [project_dir, project_dir_2, project_dir_3])
     # create remote project
-    toggle_geodiff(push_geodiff_enabled)
     shutil.copytree(TEST_DATA_DIR, project_dir)
     mc.create_project_and_push(test_project, project_dir)
 
     # make sure we have v1 also in concurrent project dirs
-    toggle_geodiff(pull_geodiff_enabled)
     mc.download_project(project, project_dir_2)
     mc.download_project(project, project_dir_3)
 
     # test push changes with diffs:
-    toggle_geodiff(push_geodiff_enabled)
     mp = MerginProject(project_dir)
     f_updated = 'base.gpkg'
     # step 1) base.gpkg updated to inserted_1_A (inserted A feature)
-    if push_geodiff_enabled:
-        shutil.move(mp.fpath(f_updated), mp.fpath_meta(f_updated))  # make local copy for changeset calculation
+    shutil.move(mp.fpath(f_updated), mp.fpath_meta(f_updated))  # make local copy for changeset calculation
     shutil.copy(mp.fpath('inserted_1_A.gpkg'), mp.fpath(f_updated))
     mc.push_project(project_dir)
-    if push_geodiff_enabled:
-        mp.geodiff.create_changeset(mp.fpath(f_updated), mp.fpath_meta(f_updated), mp.fpath_meta('push_diff'))
-        assert not mp.geodiff.has_changes(mp.fpath_meta('push_diff'))
-    else:
-        assert not os.path.exists(mp.fpath_meta(f_updated))
+    mp.geodiff.create_changeset(mp.fpath(f_updated), mp.fpath_meta(f_updated), mp.fpath_meta('push_diff'))
+    assert not mp.geodiff.has_changes(mp.fpath_meta('push_diff'))
     # step 2) base.gpkg updated to inserted_1_A_mod (modified 2 features)
-    if push_geodiff_enabled:
-        shutil.move(mp.fpath(f_updated), mp.fpath_meta(f_updated))
+    shutil.move(mp.fpath(f_updated), mp.fpath_meta(f_updated))
     shutil.copy(mp.fpath('inserted_1_A_mod.gpkg'), mp.fpath(f_updated))
     # introduce some other changes
     f_removed = 'inserted_1_B.gpkg'
@@ -343,48 +326,27 @@ def test_sync_diff(mc, push_geodiff_enabled, pull_geodiff_enabled):
     assert next((f for f in project_info['files'] if f['path'] == 'renamed.gpkg'), None)
     assert not next((f for f in project_info['files'] if f['path'] == f_removed), None)
     assert not os.path.exists(mp.fpath_meta(f_removed))
-    if push_geodiff_enabled:
-        assert 'diff' in f_remote
-        assert os.path.exists(mp.fpath_meta('renamed.gpkg'))
-    else:
-        assert 'diff' not in f_remote
-        assert not os.path.exists(mp.fpath_meta('renamed.gpkg'))
+    assert 'diff' in f_remote
+    assert os.path.exists(mp.fpath_meta('renamed.gpkg'))
 
     # pull project in different directory
-    toggle_geodiff(pull_geodiff_enabled)
     mp2 = MerginProject(project_dir_2)
     mc.pull_project(project_dir_2)
-    if pull_geodiff_enabled:
-        mp2.geodiff.create_changeset(mp.fpath(f_updated), mp2.fpath(f_updated), mp2.fpath_meta('diff'))
-        assert not mp2.geodiff.has_changes(mp2.fpath_meta('diff'))
-    else:
-        server_file_checksum = next((f['checksum'] for f in project_info['files'] if f['path'] == f_updated), '')
-        assert server_file_checksum == generate_checksum(mp2.fpath(f_updated))
+    mp2.geodiff.create_changeset(mp.fpath(f_updated), mp2.fpath(f_updated), mp2.fpath_meta('diff'))
+    assert not mp2.geodiff.has_changes(mp2.fpath_meta('diff'))
 
     # introduce conflict local change (inserted B feature to base)
     mp3 = MerginProject(project_dir_3)
     shutil.copy(mp3.fpath('inserted_1_B.gpkg'), mp3.fpath(f_updated))
     checksum = generate_checksum(mp3.fpath('inserted_1_B.gpkg'))
     mc.pull_project(project_dir_3)
-    if pull_geodiff_enabled:
-        assert not os.path.exists(mp3.fpath('base.gpkg_conflict_copy'))
-    else:
-        assert os.path.exists(mp3.fpath('base.gpkg_conflict_copy'))  #
-        assert generate_checksum(mp3.fpath('base.gpkg_conflict_copy')) == checksum
+    assert not os.path.exists(mp3.fpath('base.gpkg_conflict_copy'))
 
     # push new changes from project_3 and pull in original project
-    toggle_geodiff(push_geodiff_enabled)
     mc.push_project(project_dir_3)
-    toggle_geodiff(pull_geodiff_enabled)
     mc.pull_project(project_dir)
-    if pull_geodiff_enabled:
-        mp3.geodiff.create_changeset(mp.fpath(f_updated), mp3.fpath(f_updated), mp.fpath_meta('diff'))
-        assert not mp3.geodiff.has_changes(mp.fpath_meta('diff'))
-    else:
-        assert os.path.exists(mp.fpath('base.gpkg_conflict_copy'))
-
-    # make sure that we leave geodiff enabled for further tests
-    toggle_geodiff(True)
+    mp3.geodiff.create_changeset(mp.fpath(f_updated), mp3.fpath(f_updated), mp.fpath_meta('diff'))
+    assert not mp3.geodiff.has_changes(mp.fpath_meta('diff'))
 
 
 def test_list_of_push_changes(mc):
@@ -396,7 +358,6 @@ def test_list_of_push_changes(mc):
 
     cleanup(mc, project, [project_dir])
     shutil.copytree(TEST_DATA_DIR, project_dir)
-    toggle_geodiff(True)
     mc.create_project_and_push(test_project, project_dir)
 
     f_updated = 'base.gpkg'
@@ -452,7 +413,6 @@ def test_force_gpkg_update(mc):
     project_info = mc.project_info(project)
     assert project_info['version'] == 'v2'
     f_remote = next((f for f in project_info['files'] if f['path'] == f_updated), None)
-    assert f_remote['checksum'] == updated_checksum
     assert 'diff' not in f_remote
 
 
@@ -849,3 +809,461 @@ def test_logging(mc):
 def test_server_compatibility(mc):
     """Test server compatibility."""
     assert mc.is_server_compatible()
+
+
+def create_versioned_project(mc, project_name, project_dir, updated_file, remove=True):
+    project = API_USER + '/' + project_name
+    cleanup(mc, project, [project_dir])
+
+    # create remote project
+    shutil.copytree(TEST_DATA_DIR, project_dir)
+    mc.create_project_and_push(project_name, project_dir)
+
+    mp = MerginProject(project_dir)
+
+    # create versions 2-4
+    changes = ("inserted_1_A.gpkg", "inserted_1_A_mod.gpkg", "inserted_1_B.gpkg",)
+    for change in changes:
+        shutil.copy(mp.fpath(change), mp.fpath(updated_file))
+        mc.push_project(project_dir)
+    # create version 5 with modified file removed
+    if remove:
+        os.remove(os.path.join(project_dir, updated_file))
+        mc.push_project(project_dir)
+
+    return mp
+
+
+def test_get_versions_with_file_changes(mc):
+    """Test getting versions where the file was changed."""
+    test_project = 'test_file_modified_versions'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    f_updated = "base.gpkg"
+
+    mp = create_versioned_project(mc, test_project, project_dir, f_updated, remove=False)
+
+    project_info = mc.project_info(project)
+    assert project_info["version"] == "v4"
+    file_history = mc.project_file_history_info(project, f_updated)
+
+    with pytest.raises(ClientError) as e:
+        mod_versions = get_versions_with_file_changes(
+            mc, project, f_updated, version_from="v1", version_to="v5", file_history=file_history
+        )
+    assert "Wrong version parameters: 1-5" in str(e.value)
+    assert "Available versions: [1, 2, 3, 4]" in str(e.value)
+
+    mod_versions = get_versions_with_file_changes(
+        mc, project, f_updated, version_from="v2", version_to="v4", file_history=file_history
+    )
+    assert mod_versions == [f"v{i}" for i in range(2, 5)]
+
+
+def check_gpkg_same_content(mergin_project, gpkg_path_1, gpkg_path_2):
+    """Check if the two GeoPackages have equal content."""
+    with tempfile.TemporaryDirectory() as temp_dir:
+        diff_path = os.path.join(temp_dir, "diff_file")
+        mergin_project.geodiff.create_changeset(gpkg_path_1, gpkg_path_2, diff_path)
+        return not mergin_project.geodiff.has_changes(diff_path)
+
+
+def test_download_file(mc):
+    """Test downloading single file at specified versions."""
+    test_project = 'test_download_file'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    f_updated = "base.gpkg"
+
+    mp = create_versioned_project(mc, test_project, project_dir, f_updated)
+
+    project_info = mc.project_info(project)
+    assert project_info["version"] == "v5"
+
+    # Versioned file should have the following content at versions 2-4
+    expected_content = ("inserted_1_A.gpkg", "inserted_1_A_mod.gpkg", "inserted_1_B.gpkg")
+
+    # Download the base file at versions 2-4 and check the changes
+    f_downloaded = os.path.join(project_dir, f_updated)
+    for ver in range(2, 5):
+        mc.download_file(project_dir, f_updated, f_downloaded, version=f"v{ver}")
+        expected = os.path.join(TEST_DATA_DIR, expected_content[ver - 2])  # GeoPackage with expected content
+        assert check_gpkg_same_content(mp, f_downloaded, expected)
+
+    # make sure there will be exception raised if a file doesn't exist in the version
+    with pytest.raises(ClientError, match=f"No {f_updated} exists at version v5"):
+        mc.download_file(project_dir, f_updated, f_downloaded, version=f"v5")
+
+
+def test_download_diffs(mc):
+    """Test download diffs for a project file between specified project versions."""
+    test_project = 'test_download_diffs'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    download_dir = os.path.join(project_dir, "diffs")  # project for downloading files at various versions
+    f_updated = "base.gpkg"
+    diff_file = os.path.join(download_dir, f_updated + ".diff")
+
+    mp = create_versioned_project(mc, test_project, project_dir, f_updated, remove=False)
+
+    project_info = mc.project_info(project)
+    assert project_info["version"] == "v4"
+
+    # Download diffs of updated file between versions 1 and 2
+    mc.get_file_diff(project_dir, f_updated, diff_file, "v1", "v2")
+    assert os.path.exists(diff_file)
+    assert mp.geodiff.has_changes(diff_file)
+    assert mp.geodiff.changes_count(diff_file) == 1
+    changes_file = diff_file + ".changes1-2"
+    mp.geodiff.list_changes_summary(diff_file, changes_file)
+    with open(changes_file, 'r') as f:
+        changes = json.loads(f.read())["geodiff_summary"][0]
+        assert changes["insert"] == 1
+        assert changes["update"] == 0
+
+    # Download diffs of updated file between versions 2 and 4
+    mc.get_file_diff(project_dir, f_updated, diff_file, "v2", "v4")
+    changes_file = diff_file + ".changes2-4"
+    mp.geodiff.list_changes_summary(diff_file, changes_file)
+    with open(changes_file, 'r') as f:
+        changes = json.loads(f.read())["geodiff_summary"][0]
+        assert changes["insert"] == 0
+        assert changes["update"] == 1
+
+    with pytest.raises(ClientError) as e:
+        mc.get_file_diff(project_dir, f_updated, diff_file, "v4", "v1")
+    assert "Wrong version parameters" in str(e.value)
+    assert "version_from needs to be smaller than version_to" in str(e.value)
+
+    with pytest.raises(ClientError) as e:
+        mc.get_file_diff(project_dir, f_updated, diff_file, "v4", "v5")
+    assert "Wrong version parameters" in str(e.value)
+    assert "Available versions: [1, 2, 3, 4]" in str(e.value)
+
+
+def _use_wal(db_file):
+    """ Ensures that sqlite database is using WAL journal mode """
+    con = sqlite3.connect(db_file)
+    cursor = con.cursor()
+    cursor.execute('PRAGMA journal_mode=wal;')
+
+
+def _create_test_table(db_file):
+    """ Creates a table called 'test' in sqlite database. Useful to simulate change of database schema. """
+    con = sqlite3.connect(db_file)
+    cursor = con.cursor()
+    cursor.execute('CREATE TABLE test (fid SERIAL, txt TEXT);')
+    cursor.execute('INSERT INTO test VALUES (123, \'hello\');')
+    cursor.execute('COMMIT;')
+
+
+def _check_test_table(db_file):
+    """ Checks whether the 'test' table exists and has one row - otherwise fails with an exception. """
+    #con_verify = sqlite3.connect(db_file)
+    #cursor_verify = con_verify.cursor()
+    #cursor_verify.execute('select count(*) from test;')
+    #assert cursor_verify.fetchone()[0] == 1
+    assert _get_table_row_count(db_file, 'test') == 1
+
+
+def _get_table_row_count(db_file, table):
+    con_verify = sqlite3.connect(db_file)
+    cursor_verify = con_verify.cursor()
+    cursor_verify.execute('select count(*) from {};'.format(table))
+    return cursor_verify.fetchone()[0]
+
+
+def _is_file_updated(filename, changes_dict):
+    """ checks whether a file is listed among updated files (for pull or push changes) """
+    for f in changes_dict['updated']:
+        if f['path'] == filename:
+            return True
+    return False
+
+
+class AnotherSqliteConn:
+    """ This simulates another app (e.g. QGIS) having a connection open, potentially
+    with some active reader/writer.
+
+    Note: we use a subprocess here instead of just using sqlite3 module from python
+    because pygeodiff and python's sqlite3 module have their own sqlite libraries,
+    and this does not work well when they are used in a single process. But if we
+    use another process, things are fine. This is a limitation of how we package
+    pygeodiff currently.
+    """
+    def __init__(self, filename):
+        self.proc = subprocess.Popen(
+            ['python3', os.path.join(os.path.dirname(__file__), 'sqlite_con.py'), filename],
+            stdin=subprocess.PIPE,
+            stderr=subprocess.PIPE)
+
+    def run(self, cmd):
+        self.proc.stdin.write(cmd.encode()+b'\n')
+        self.proc.stdin.flush()
+
+    def close(self):
+        out,err = self.proc.communicate(b'stop\n')
+        if self.proc.returncode != 0:
+            raise ValueError("subprocess error:\n" + err.decode('utf-8'))
+
+
+def test_push_gpkg_schema_change(mc):
+    """ Test that changes in GPKG get picked up if there were recent changes to it by another
+    client and at the same time geodiff fails to find changes (a new table is added)
+    """
+
+    test_project = 'test_push_gpkg_schema_change'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    project_dir_verify = os.path.join(TMP_DIR, test_project + '_verify')
+    test_gpkg_verify = os.path.join(project_dir_verify, 'test.gpkg')
+
+    cleanup(mc, project, [project_dir, project_dir_verify])
+    # create remote project
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    #shutil.copytree(TEST_DATA_DIR, project_dir)
+    mc.create_project_and_push(test_project, project_dir)
+
+    mp = MerginProject(project_dir)
+
+    mp.log.info(' // create changeset')
+
+    mp.geodiff.create_changeset(mp.fpath('test.gpkg'), mp.fpath_meta('test.gpkg'), mp.fpath_meta('diff-0'))
+
+    mp.log.info(' // use wal')
+
+    _use_wal(test_gpkg)
+
+    mp.log.info(' // make changes to DB')
+
+    # open a connection and keep it open (qgis does this with a pool of connections too)
+    acon2 = AnotherSqliteConn(test_gpkg)
+    acon2.run('select count(*) from simple;')
+
+    # add a new table to ensure that geodiff will fail due to unsupported change
+    # (this simulates an independent reader/writer like GDAL)
+    _create_test_table(test_gpkg)
+
+    _check_test_table(test_gpkg)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_basefile)
+
+    mp.log.info(' // create changeset (2)')
+
+    # why already here there is wal recovery - it could be because of two sqlite libs linked in one executable
+    # INDEED THAT WAS THE PROBLEM, now running geodiff 1.0 with shared sqlite lib seems to work fine.
+    with pytest.raises(pygeodiff.geodifflib.GeoDiffLibError):
+        mp.geodiff.create_changeset(mp.fpath('test.gpkg'), mp.fpath_meta('test.gpkg'), mp.fpath_meta('diff-1'))
+
+    _check_test_table(test_gpkg)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_basefile)
+
+    mp.log.info(' // push project')
+
+    # push pending changes (it should include addition of the new table)
+    # at this point we still have an open sqlite connection to the GPKG, so checkpointing will not work correctly)
+    mc.push_project(project_dir)
+
+    # WITH TWO SQLITE copies: fails here  (sqlite3.OperationalError: disk I/O error)  + in geodiff log: SQLITE3: (283)recovered N frames from WAL file
+    _check_test_table(test_gpkg)
+
+    # OLD: fails here when con2 is still alive: checkpointing fails and basefile has incorrect value
+    _check_test_table(test_gpkg_basefile)
+
+    # download the project to a new directory and verify the change was pushed correctly
+    mc.download_project(project, project_dir_verify)
+
+    # OLD: fails here
+    _check_test_table(test_gpkg_verify)
+
+    acon2.close()
+
+
+@pytest.mark.parametrize("extra_connection", [False, True])
+def test_rebase_local_schema_change(mc, extra_connection):
+    """
+    Checks whether a pull with failed rebase (due to local DB schema change) is handled correctly,
+    i.e. a conflict file is created with the content of the local changes.
+    """
+
+    test_project = 'test_rebase_local_schema_change'
+    if extra_connection:
+        test_project += '_extra_conn'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir
+    project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    cleanup(mc, project, [project_dir, project_dir_2])
+
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL, that's the more common and more difficult scenario
+    mc.create_project_and_push(test_project, project_dir)
+
+    if extra_connection:
+        # open a connection and keep it open (qgis does this with a pool of connections too)
+        con_extra = sqlite3.connect(test_gpkg)
+        cursor_extra = con_extra.cursor()
+        cursor_extra.execute('select count(*) from simple;')
+
+    # Download project to the concurrent dir + add a feature + push a new version
+    mc.download_project(project, project_dir_2)  # download project to concurrent dir
+    mp_2 = MerginProject(project_dir_2)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_A.gpkg'), mp_2.fpath('test.gpkg'))
+    mc.push_project(project_dir_2)
+
+    # Change schema in the primary project dir
+    _create_test_table(test_gpkg)
+
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert _is_file_updated('test.gpkg', pull_changes)
+    assert _is_file_updated('test.gpkg', push_changes)
+
+    assert not os.path.exists(test_gpkg_conflict)
+
+    mc.pull_project(project_dir)
+
+    assert os.path.exists(test_gpkg_conflict)
+
+    # check the results after pull:
+    # - conflict copy should contain the new table
+    # - local file + basefile should not contain the new table
+
+    _check_test_table(test_gpkg_conflict)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_basefile)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg)
+
+    # check that the local file + basefile contain the new row, and the conflict copy doesn't
+
+    assert _get_table_row_count(test_gpkg, 'simple') == 4
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 4
+    assert _get_table_row_count(test_gpkg_conflict, 'simple') == 3
+
+
+@pytest.mark.parametrize("extra_connection", [False, True])
+def test_rebase_remote_schema_change(mc, extra_connection):
+    """
+    Checks whether a pull with failed rebase (due to remote DB schema change) is handled correctly,
+    i.e. a conflict file is created with the content of the local changes.
+    """
+
+    test_project = 'test_rebase_remote_schema_change'
+    if extra_connection:
+        test_project += '_extra_conn'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir
+    project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    cleanup(mc, project, [project_dir, project_dir_2])
+
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL, that's the more common and more difficult scenario
+    mc.create_project_and_push(test_project, project_dir)
+
+    # Download project to the concurrent dir + change DB schema + push a new version
+    mc.download_project(project, project_dir_2)
+    _create_test_table(test_gpkg_2)
+    mc.push_project(project_dir_2)
+
+    # do changes in the local DB (added a row)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_A.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL
+
+    if extra_connection:
+        # open a connection and keep it open (qgis does this with a pool of connections too)
+        con_extra = sqlite3.connect(test_gpkg)
+        cursor_extra = con_extra.cursor()
+        cursor_extra.execute('select count(*) from simple;')
+
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert _is_file_updated('test.gpkg', pull_changes)
+    assert _is_file_updated('test.gpkg', push_changes)
+
+    assert not os.path.exists(test_gpkg_conflict)
+
+    mc.pull_project(project_dir)
+
+    assert os.path.exists(test_gpkg_conflict)
+
+    # check the results after pull:
+    # - conflict copy should not contain the new table
+    # - local file + basefile should contain the new table
+
+    _check_test_table(test_gpkg)
+    _check_test_table(test_gpkg_basefile)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_conflict)
+
+    # check that the local file + basefile don't contain the new row, and the conflict copy does
+
+    assert _get_table_row_count(test_gpkg, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_conflict, 'simple') == 4
+
+
+@pytest.mark.parametrize("extra_connection", [False, True])
+def test_rebase_success(mc, extra_connection):
+    """
+    Checks whether a pull with successful rebase is handled correctly.
+    i.e. changes are merged together and no conflict files are created.
+    """
+
+    test_project = 'test_rebase_success'
+    if extra_connection:
+        test_project += '_extra_conn'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir
+    project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    cleanup(mc, project, [project_dir, project_dir_2])
+
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL, that's the more common and more difficult scenario
+    mc.create_project_and_push(test_project, project_dir)
+
+    # Download project to the concurrent dir + add a row + push a new version
+    mc.download_project(project, project_dir_2)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_A.gpkg'), test_gpkg_2)
+    _use_wal(test_gpkg)  # make sure we use WAL
+    mc.push_project(project_dir_2)
+
+    # do changes in the local DB (added a row)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_B.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL
+
+    if extra_connection:
+        # open a connection and keep it open (qgis does this with a pool of connections too)
+        con_extra = sqlite3.connect(test_gpkg)
+        cursor_extra = con_extra.cursor()
+        cursor_extra.execute('select count(*) from simple;')
+
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert _is_file_updated('test.gpkg', pull_changes)
+    assert _is_file_updated('test.gpkg', push_changes)
+
+    assert not os.path.exists(test_gpkg_conflict)
+
+    mc.pull_project(project_dir)
+
+    assert not os.path.exists(test_gpkg_conflict)
+
+    # check that the local file + basefile don't contain the new row, and the conflict copy does
+
+    assert _get_table_row_count(test_gpkg, 'simple') == 5
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 4
